from OpenGL.GL import *
from OpenGL.GL.EXT.texture_edge_clamp import *
import math
import pygame
from vector import Vector
import data

R = 2
B = 0.5
S = 0.6

class Font:
    def __init__(self):
        self.texture = 0

    def load(self, name):
        img = pygame.image.load(name)
        glEnable(GL_TEXTURE_2D)
        self.texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.texture)
        pixels = pygame.image.tostring(img, "RGBA", True)
        glTexImage2D(GL_TEXTURE_2D, 0, 4, 256, 256, 0,
            GL_RGBA, GL_UNSIGNED_BYTE,
            pixels)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)

    def write(self, px, py, ps, text, center = True):
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glBindTexture(GL_TEXTURE_2D, font.texture)

        if center: px -= len(text) * ps * 0.5

        if not self.texture: self.load(data.filepath("tiny.png"))
        glBegin(GL_QUADS)
        for letter in text:
            i = ord(letter) - 32
            s = 6 / 256.0
            ix = i % 16
            iy = i / 16
            x = (50 + ix * 10) / 256.0
            y = (256 - 99 - iy * 10 - 6) / 256.0

            glTexCoord2f(x, y)
            glVertex2f(px, py - ps)
            glTexCoord2f(x + s, y)
            glVertex2f(px + ps, py - ps)
            glTexCoord2f(x + s, y + s)
            glVertex2f(px + ps, py)
            glTexCoord2f(x, y + s)
            glVertex2f(px, py)
            px += ps
        glEnd()

        glDisable(GL_TEXTURE_2D)

font = Font()

def arc(x, y, r, sa, ea):
        n = int(max(1, r * 1000))
        for i in range(n + 1):
            a = sa + i * (ea - sa) / n
            glVertex2f(x + math.cos(a) * r, y + math.sin(a) * r)

def circle(x, y, r):
        n = int(max(1, r * 1000))
        for i in range(n + 1):
            a = i * 2 * math.pi / n
            glVertex2f(x + math.cos(a) * r, y + math.sin(a) * r)

def intersect(A, n, B, l):
    """
    Returns which multiple of l to add to B so the result lies on the plane
    defined by containing point A and normal n.
    """
    # How it works
    # We have a plane: It contains A and has a normal of n.
    # And we have a line: It starts in B and goes along l.
    # Where do they intersect?
    # 1. (? - A) dot n = 0 [? must be on the plane]
    # 2. ? = B + t * l [? must be on the line]
    #    [2 in 1 gives:]
    # 3. (B + t * l - A) dot n = 0
    # -> (B + t * l) dot n = A dot n
    # -> B dot n + t * l dot n = A dot n
    #    t * l dot n = A dot n - B dot n
    #    t = (A - B) dot n / l dot n
    d = l.dot(n)
    t = (A - B).dot(n)
    return t / d

def render_horizon(self, cam):
    # How to find the artificial horizon:
    #
    # Assume two infinite planes. One is all green and at the bottom of the
    # world, one is all blue and at the top. Now, wherever we are and
    # wherever we look, the whole screen will be filled with blue and
    # green. If we look straight up, it will be all blue, if we look
    # straight down, all green. If we look perfectly horizontal, the upper
    # part will be blue, the lower green.
    #
    # Basically, if the ray from the viewpoint to a screen pixel goes down
    # in world space, the pixel is green. If it goes up, it is blue.
    #
    # The artificial horizon is just the line where the y component of view
    # rays in world coordinates is 0.

    f = 0.002
    r = R * 10
    x = cam.w - (R + B) * 10 * f
    y = cam.h - (2 * (R + R + S) + R + S) * 10 * f

    cr = self.r
    cu = self.u
    cb = self.b
    cx = Vector(0.0, 1.0, 0.0)

    # get view direction (without up/down)
    compass = -cb
    compass.y = 0.0

    # get another point on the horizon
    s = compass + compass.cross(cx)

    # intersect both with the view plane
    V = self.position - cb
    t1 = intersect(V, -cb, self.position, compass)
    T1 = self.position + compass * t1
    t2 = intersect(V, -cb, self.position, s)
    T2 = self.position + s * t2

    # T1 and T2 are two points, in world space, through which the horizon line
    # runs.
    T1 -= V
    T2 -= V

    # get the offsets along right and up vectors (i.e. project to camera space).
    O = Vector(T1.dot(cr), T1.dot(cu), 0)
    T = Vector(T2.dot(cr) - O.x, T2.dot(cu) - O.y, 0)
    O *= r
    T *= r

    # intersect with a circle
    # x * x + y * y = r * r
    # Ox + u * Tx = x
    # Oy + u * Ty = y
    # OxOx + uu * TxTx + 2 * u * Tx + OyOy + uu * TyTy + 2 * u * Ty = rr
    # a = TxTx + TyTy
    # b = 2 * (Tx + Ty)
    # c = OxOx + OyOy - rr
    # a * u^1 + b * u + c = 0
    # D = b^2 - 4 ac
    # u = (-b +- D^0.5) / (2a)

    glColor4f(0.4, 0.8, 1, 0.5)
    glBegin(GL_POLYGON)
    circle(x, y, r * f)
    glEnd()

    # eww, numerically unstable :P
    a = T.x * T.x + T.y * T.y
    b = 2 * (T.x + T.y)
    c = O.x * O.x + O.y * O.y - r * r
    D = b * b - 4 * a * c
    if D > 0: # horizon visible
        u1 = (-b + (D ** 0.5)) / (2 * a)
        u2 = (-b - (D ** 0.5)) / (2 * a)

        x1 = O.x + u1 * T.x
        y1 = O.y + u1 * T.y
        x2 = O.x + u2 * T.x
        y2 = O.y + u2 * T.y

        sa = math.atan2(y2, x2)
        ea = math.atan2(y1, x1)
        if ea < sa: ea += 2 * math.pi
        glColor4f(0, 0, 0.5, 0.5)
        glBegin(GL_POLYGON)
        arc(x, y, r * f, sa, ea)
        glEnd()

        glBegin(GL_LINES)
        glColor4f(1, 1, 1, 1)
        glVertex2f(x + x1 * f, y + y1 * f)
        glVertex2f(x + x2 * f, y + y2 * f)
        glEnd()
    elif self.b.y > 0:
        glColor4f(0, 0, 0.5, 0.5)
        glBegin(GL_POLYGON)
        circle(x, y, r * f)
        glEnd()

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r * f)
    glEnd()

def render_compass(self, target, cam):
    # This is the orientation in the world x/z plane.
    f = 0.02
    r = R * f
    x = cam.w - (R + B) * f
    y = cam.h - (R + S) * f
    compass = -self.b.copy() * r

    glColor3f(1, 1, 1)
    font.write(x, y + r * 0.9, f * 0.7, "N")
    font.write(x, y - r * 0.9 + f * 0.7, f * 0.7, "S")
    font.write(x + r * 0.8, y + f * 0.35, f * 0.7, "E")
    font.write(x - r * 0.8, y + f * 0.35, f * 0.7, "W")

    glBegin(GL_LINES)
    glColor3f(1, 1, 1)
    glVertex2f(x, y)
    glVertex2f(x + compass.x, y - compass.z)
    glEnd()

    if target:
        t = target.position - self.position
        t.y = 0
        t = t.normalize()
        glBegin(GL_LINES)
        glColor3f(1, 0, 0)
        glVertex2f(x, y)
        glVertex2f(x + t.x * r, y - t.z * r)
        glEnd()

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r)
    glEnd()

def render_pitch(self, target, cam):
    # The up/down orientation. Also considers if the self is upside down.
    up = Vector(0, 1, 0)
    c = self.b.dot(up)
    s = self.b.cross(up).norm()
    if self.u.dot(up) < 0:
        s = -s
    f = 0.02
    r = R * f
    x = cam.w - (R + B) * f
    y = cam.h - (1 * (R + R + S) + R + S) * f

    glColor3f(1, 1, 1)
    font.write(x, y + r * 0.9, f * 0.7, "up")
    font.write(x, y - r * 0.9 + f * 0.7, f * 0.7, "down")

    glBegin(GL_LINES)
    glColor3f(1, 1, 1)
    glVertex2f(x, y)
    glVertex2f(x - s * r, y - c * r)
    glEnd()

    if target:
        t = target.position - self.position
        l = t.norm()
        if l > 0.1:
            s = t.y / l
            glBegin(GL_LINES)
            xr = math.sqrt(1 - s * s) * r
            glColor3f(1, 0, 0)
            glVertex2f(x - xr, y + s * r)
            glVertex2f(x + xr, y + s * r)
            glEnd()

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r)
    glEnd()

def render_rpm(self, cam):
    # Render engine RPM.
    f = 0.02
    r = R * f
    x = cam.w - (R + B) * f
    y = cam.h - (3 * (R + R + S) + R + S) * f
    v = self.engine / self.ENGINE_MAX
    a = math.pi + math.pi * v
    vx = math.cos(a) * r
    vy = math.sin(a) * r

    glColor4f(v, -v, 0, 0.6)
    glBegin(GL_POLYGON)
    glVertex2f(x, y)
    arc(x, y, r, math.pi, a)
    glEnd()

    glBegin(GL_LINES)
    glColor3f(1, 1, 1)
    glVertex2f(x, y)
    glVertex2f(x - r, y)
    glVertex2f(x, y)
    glVertex2f(x + vx, y + vy)
    glEnd()

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r)
    glEnd()

    glColor3f(1, 1, 1)
    font.write(x, y, f * 0.7, "RPM")

def render_hand(self, x, y, a, r):
    vx = math.cos(a) * r
    vy = math.sin(a) * r
    glBegin(GL_LINES)
    glColor3f(1, 1, 1)
    glVertex2f(x, y)
    glVertex2f(x + vx, y + vy)
    glEnd()

def render_altimeter(self, cam):
    f = 0.02
    r = R * f
    x = cam.w - (R + B) * f
    y = cam.h - (4 * (R + R + S) + R + S) * f

    v = self.position.y / 20000
    a = math.pi / 2 - math.pi * 2 * v
    render_hand(self, x, y, a, r * 0.5)

    v = self.position.y / 1000 - int(self.position.y / 1000)
    a = math.pi / 2 - math.pi * 2 * v
    render_hand(self, x, y, a, r * 0.75)

    v = self.position.y / 100 - int(self.position.y / 100)
    a = math.pi / 2 - math.pi * 2 * v
    render_hand(self, x, y, a, r)

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r)
    glEnd()

    font.write(x, y, f * 0.5, "%d" % (20000 - self.position.y))
    font.write(x, y - f * 0.5, f * 0.5, "ALT")

def render_air(self, cam):
    f = 0.02
    r = R * f
    x = cam.w - (R + B) * f
    y = cam.h - (5 * (R + R + S) + R + S) * f

    v = self.air / 3600.0
    if v > 1: v = 1
    if v < 0: v = 0
    a = math.pi * -0.75 - math.pi * 1.5 * v
    render_hand(self, x, y, a, r)

    glColor4f(1, 1, 1, 1)
    glBegin(GL_LINE_STRIP)
    circle(x, y, r)
    glEnd()

    glColor3f(1, 1, 1)
    font.write(x, y, f * 0.5, "OXYGEN")

def render_score(self, cam):
    f = 0.02
    r = R * f
    x = cam.w - (R + R + B) * f
    y = -cam.h + 1.4 * S * f

    font.write(x, y, f * 0.5, "X=%d Y=%d" % (self.position.x, self.position.z))
    y += f * 0.75
    font.write(x, y, f * 0.5, "Buoys left: %02d" % self.buoys_left)
    y += f * 0.75
    font.write(x, y, f * 0.5, "Score: %07d" % self.score)

def render_help(self, cam):
    f = 0.02
    x = -cam.w
    y = cam.h
    text = """
F1.........toggle help
F10........toggle mouse capture
Enter......toggle fullscreen
Esc........exit
Mouse......move camera
M. wheel...zoom
Left M
or Shift...accelerate
Rght M 
or Ctrl....decelerate
Left or A..rotate left
Rght or D..rotate right
Up or W....rotate uo
Down or S..rotate down
"""
    for line in text.splitlines():
        font.write(x, y, f * 0.6, line, center = False)
        y -= f * 0.6

def render_level(level, hint, cam):
    f = 0.02
    x = 0
    y = f * 2
    glColor3f(1, 1, 1)
    font.write(x, y, f * 2, "Level %d" % level)
    y -= f * 2
    glColor3f(1, 1, 0)
    font.write(x, y, f * 0.7, hint)
