import math, quaternion

class Vector:
    """
    A simple 3D vector, with some extra useful stuff for quaternions,
    transformations and rotation.
    """
    def __init__(v, x, y, z):
        v.x = x
        v.y = y
        v.z = z

    def copy(v):
        """
        Return a copy of the vector.
        """
        return Vector(v.x, v.y, v.z)

    def __iadd__(v, w):
        v.x += w.x
        v.y += w.y
        v.z += w.z
        return v

    def __isub__(v, w):
        v.x -= w.x
        v.y -= w.y
        v.z -= w.z
        return v

    def __imul__(v, s):
        v.x *= s
        v.y *= s
        v.z *= s
        return v

    def __idiv__(v, s):
        v.x /= s
        v.y /= s
        v.z /= s
        return v

    def __neg__(v):
        return Vector(-v.x, -v.y, -v.z)

    def __mul__(v, s):
        r = Vector(v.x, v.y, v.z)
        r *= s
        return r

    def __div__(v, s):
        r = Vector(v.x, v.y, v.z)
        r /= s
        return r

    def __add__(v, s):
        r = Vector(v.x, v.y, v.z)
        r += s
        return r

    def __sub__(v, s):
        r = Vector(v.x, v.y, v.z)
        r -= s
        return r

    def dot(v, w):
        """
        The dot product is a number. The number corresponds to the cosine
        between the two vectors times their lengths. So the angle between the
        vectors would be: angle = acos(v . w / (|v| * |w|)). If the dot product
        is 0, the two vectors conversely are orthogonal. The sign can be used to
        determine which side of a plane a point is on.
        """
        return v.x * w.x + v.y * w.y + v.z * w.z

    def cross(v, w):
        """
        The cross product results in a vector orthogonal to both v and w. The
        length of the resulting vector corresponds to the sine of the angle
        between the two vectors and their lengths. So the angle between the
        vectors would be: angle = asin(|v x w| / (|v| * |w|)). If the cross
        product is the 0 vector, the two input vectors are parallel.
        """ 
        x = v.y * w.z - w.y * v.z
        y = v.z * w.x - w.z * v.x
        z = v.x * w.y - w.x * v.y
        return Vector(x, y, z)

    def norm(v):
        """
        Return the norm of the vector.
        """
        return math.sqrt(v.dot(v))

    def normalize(v):
        """
        Return a normalized version of the vector.
        """
        return v / v.norm()

    def quatmul(v, q):
        """
        Multiply the vector with a quaternion. The result is a quaternion. For
        example if your vector is a rotation, the resulting quaternion will be a
        quaternion who rotates whatever it did plus this additional rotation.
        """
        w = - v.x * q.x - v.y * q.y - v.z * q.z
        x =   v.x * q.w + v.y * q.z - v.z * q.y
        y =   v.y * q.w + v.z * q.x - v.x * q.z
        z =   v.z * q.w + v.x * q.y - v.y * q.x
        return quaternion.Quaternion(w, x, y, z)

    def transform(v, p, r, u, b):
        """
        Return a new vector obtained by transforming this vector by a coordinate
        system with the given origin and given right/up/back vectors. This is
        used if the vector is in world coordinates, and you want to transform it
        to camera coordinates, where p/r/u/b define camera position and
        orientation.
        """
        w = v - p
        x = w.dot(r)
        y = w.dot(u)
        z = w.dot(b)
        return Vector(x, y, z)

    def backtransform(v, p, r, u, b):
        """
        Do the inverse of transform, i.e. you can use it to transform from
        camera back to world coordinates.
        """
        x = r * v.x
        y = u * v.y
        z = b * v.z
        return p + x + y + z

    def rotate(v, a, angle):
        """
        Rotate the vector around axis a by angle in counter clockwise direction.
        If this vector is a point in world space, then the axis of rotation is
        defined by the origin and the a vector.
        """
        c = math.cos(angle)
        s = math.sin(angle)

        r = a * a.x * (1 - c)
        u = a * a.y * (1 - c)
        b = a * a.z * (1 - c)

        r.x += c
        r.y += a.z * s
        r.z -= a.y * s

        u.x -= a.z * s
        u.y += c
        u.z += a.x * s

        b.x += a.y * s
        b.y -= a.x * s
        b.z += c

        x = v.dot(r)
        y = v.dot(u)
        z = v.dot(b)

        return Vector(x, y, z)

    def __getitem__(v, i):
        """Allow array access to the vector's members."""
        return getattr(v, ["x", "y", "z"][i])

    def __repr__(v):
        return "Vector(%.1f, %.1f, %.1f)" % (v.x, v.y, v.z)

